
# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app),
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

module Hermes
  # Used for enterprise and enterprises sub-entities
  class Entity
    # tous les champs qui sont non décrits ou avec absence de connaissance fonctionnel ont été
    # tirés des XSD et / ou XML des données d'origines

    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::AlsIdConcern
    include Hermes::Concerns::Status
    include ::Concerns::Raci
    include ::Concerns::NotificationCenter::Targetable

    hermes_tools 'label_interpreter'

    # champs de dépense de points à prendre en compte au before_update
    SPENT_ON_FIELDS = %w[
      spent_on_completed_requests
      spent_on_ongoing_requests
      spent_on_other
    ].freeze

    NOT_NULL_INTEGER_FIELDS = SPENT_ON_FIELDS + %w[
      annual_stock
      total_spent
      available_points
    ]

    field :test_created_at, type: DateTime

    field :title, type: String
    field :siret, type: String
    field :siren, type: String

    field :manual_mode, type: Boolean, default: false 
    field :is_master, type: Boolean, default: false 

    field :is_root, type: Boolean, default: true
    field :root_id, type: BSON::ObjectId                                # censé être utilisé pour identifier l'entity comme racine de parenté
    field :parent_id, type: BSON::ObjectId                              # censé être utilisé pour identifier le parent direct d'une entity
    field :id_rex, type: String                                         # id rex correspondante
    field :id_cristal, type: String                                     # id cristal correspondante
    field :id_parent_cristal, type: String                              # id cristal parent

    field :auto_validation, type: Boolean, default: false               # permet de valider de manière automatique au non les informations de EntityApplicant
    field :auto_request_validation, type: Boolean, default: true        # permet de valider automatiquement les demandes dès que l'entity_id change sur la demande

    field :priorities_enabled, type: Boolean, default: false            # non décrit, en rapport avec la gestion des points de priorité je supposes

    field :in_peec, type: Boolean, default: false                       # non décrit, censé être utilisé pour identifier les entreprises associé au programme peec
    field :achieved_goal, type: Boolean                                 # non décrit, censé gérer l'atteinte ou non des objectifs de l'entreprise

    field :spent_on_completed_requests, type: Integer, default: 0       # permet de stocker et suivre les points de priorité dépensés sur les demandes finalisées
    field :spent_on_ongoing_requests, type: Integer, default: 0         # permet de stocker et suivre les points de priorité dépensés sur les demandes en cours de réalisation
    field :spent_on_other, type: Integer, default: 0                    # ajouté pour identifier un mai-fonctionnement, tout point dépensé qui est assigné à ce champ
                                                                        # indique un problème de gestion des status sur la demande concernée

    field :annual_stock, type: Integer, default: 0                   # stock de points de priorité annuels de l'entreprise (aucune gestion de la réinitialisation car non décrite)
    field :total_spent, type: Integer, default: 0                       # total dépensé (somme des trois champs spent_on)
    field :available_points, type: Integer, default: 0                    # balance des points de priorités restant à l'entreprise

    field :supervisor_group_id, type: BSON::ObjectId                    # group superviseur (lecture des objets associés - demandes / offres) de l'entreprise (auto-généré)
    field :manager_group_id, type: BSON::ObjectId                       # group manageur (écriture des objets associés - demandes / offres) de l'entreprise (auto-généré)

    field :cre_email, type: String                                      # email de l'utilisateur cre
    field :cre_first_name, type: String                                          # prenom de l'utilisateur cre
    field :cre_last_name, type: String                                          # prenom de l'utilisateur cre

    field :cre_national_id_cristal, type: String
    field :cre_national_email, type: String
    field :cre_national_first_name, type: String
    field :cre_national_last_name, type: String

    #has_one :image, class_name: '::Doc'
    #has_one :contact_group, class_name: 'Hermes::ContactGroup'
    #has_and_belongs_to_many :housing_offers, class_name: 'Hermes::HousingOffer'
    # TODO: handle also with housing_offer_ids=
    has_many :form_levels, class_name: 'Hermes::HousingRequestEntityFormLevel'
    belongs_to(:lessor, class_name: 'Hermes::Lessor', index: true)
    belongs_to(:group, class_name: 'Group', index: true)

    common_indexes *%i[
      title
      siret
      siren
      manual_mode
      is_root
      root_id
      parent_id
      id_rex
      id_cristal
      id_parent_cristal
      auto_validation
      auto_request_validation
      priorities_enabled
      in_peec
      achieved_goal
      spent_on_completed_requests
      spent_on_ongoing_requests
      spent_on_other
      annual_stock
      total_spent
      available_points
      supervisor_group_id
      manager_group_id
      cre_email
      cre_first_name
      cre_last_name
    ]

    before_create :before_create_actions
    before_update :before_actions # https://dev.hermes.openstack.local/admin/families/5d3ea80c4ac06e63808d0fc5/mapping
    before_save :before_save_actions
    after_save :after_actions
    after_destroy :after_destroy_actions

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    validates :siret, length: { is: 14 }, allow_blank: true
    validates :siren, length: { is: 9 }, allow_blank: true
    Entity.add_status(:single)
    # def doc_id
    #   image && image.id
    # end

    # ===============
    # UTILITY METHODS
    # ===============

    # utilitaire pour obtenir le modèle adéquat avec une étiquette
    # def target_class(name = 'entity')
    #   raise ArgumentError, 'parameter can\'t be blank' if name.blank?

    #   case name.downcase
    #   when 'entity'
    #     Hermes::Entity
    #   when 'lessor'
    #     Hermes::Lessor
    #   else
    #     raise ArgumentError, "'#{name}' is not a valid class label - use 'entity' or 'lessor'"
    #   end
    # end

    # récupération de la racine de l'arborescence en fonction du root_id
    def root(target_class_name = 'entity')
      return unless root_id.present?

      target_klass = target_class(target_class_name)
      target_klass.find_by(id: root_id)
    end

    def self.not_null_integer_fields
      NOT_NULL_INTEGER_FIELDS
    end 

    def direct_children
      Hermes::Entity.where(parent_id: id)
    end

    def responsible
      if self.user.nil?
        responsible_id = "1"
      else 
        responsible_id = self.user.id
      end 
      return responsible_id
    end 

    def find_level(entity = self, i = 0)
      return i if entity.is_root
      find_level(entity.parent, i +1)
    end 

    # récupération du parent de l'élément en fonction du parent_id
    def parent(depth_level = 1, target_class_name = 'entity', return_iteration = false)
      return if is_root == true
      return unless parent_id.present?
      raise ArgumentError, 'depth_level (first parameter) must be a number' unless depth_level.is_a?(Integer)
      raise ArgumentError, 'depth_level == 0 is the entity'                 if depth_level == 0

      # compte le nombre de pas produit pour atteindre le résultat
      step = 0
      found_parent =  if depth_level == -1
                        if root_id.present? && return_iteration != true
                          root(target_class_name)
                        else
                          method_result = find_root(target_class_name, return_iteration)
                          step = method_result[:iteration_quantity]
                          method_result[:object]
                        end
                      elsif depth_level == 1
                        step += 1
                        target_klass = target_class(target_class_name)
                        target_klass.find_by(id: parent_id)
                      elsif depth_level > 1 && parent.present?
                        previous_step_parent = nil
                        step_parent = parent
                        step = 1
                        while step != depth_level && step_parent.present?
                          previous_step_parent = step_parent
                          step_parent = step_parent.parent
                          step += 1
                        end
                        raise ArgumentError, "Requested parent at #{depth_level} depth, but depth stops at #{step}" unless step == depth_level

                        step_parent
                      else
                        raise ArgumentError, "#{depth_level} is neither positive or equal to -1"
                      end
      raise ArgumentError, "No parent found at depth #{depth_level}" if found_parent.blank?

      # gestion du type d'objet retourné en fonction du paramètre @target_class_name
      found_parent = found_parent.lessor_instance if target_class_name == 'lessor'
      # gestion du format de retour en fonction du paramètre @return_iteration
      return { object: found_parent, iteration_quantity: step } if return_iteration == true

      found_parent
    end

    # retourne l'arboresence depuis, jusqu'à
    def tree(
      from_parent_depth = -1,               # niveau du parent depuis lequel générer l'arborescence ; -1 == root
      depth_from_target = -1,               # niveau de profondeur, depuis le parent, jusqu'à laquelle générer l'arborescence ; -1 == profondeur maximale
      target_class_name = 'entity',         # restreint le retour au type de model choisit
      for_user = nil,
      use_acl = false
    )
      raise ArgumentError, "from_parent_depth parameter must be '>= 0' || '== -1'" unless from_parent_depth == -1 || from_parent_depth >= 0

      from_parent = if is_root == true || from_parent_depth == 0
                      from_parent_depth = 0
                      self
                    elsif from_parent_depth == -1 || from_parent_depth > 0
                      method_result = parent(from_parent_depth, target_class_name, true)
                      if method_result.blank?
                        from_parent_depth = method_result[:iteration_quantity]
                        method_result[:object]
                      end
                    end
      raise ArgumentError, "no parent found at depth #{from_parent_depth}" if from_parent.blank?

      seeked_depth = if depth_from_target < 0
                       nil
                     else
                       from_parent_depth + depth_from_target
                     end

      self.class.rex_aggregation(from_parent.id, target_class_name, seeked_depth, for_user, use_acl).first
    end

    def atree
      target_klass = target_class(target_class_name)

      target_klass


      agg=Hermes::Entity.collection.aggregate(
        [
          { "$match" => { "parent_id"=> nil } },
          { "$project" => { "_id": 1, "title": 1, "parent_id": 1} },
          { "$graphLookup" => {
            "from"=> "hermes_entities",
            "startWith"=> "$_id",
            "connectFromField"=> "_id",
            "connectToField"=> "parent_id",
            "as"=> "hermes_entities"
          }},
          #{ "$addFields"=> {
          #  "entities"=> {
          #    "$reverseArray"=> "$entities"
          #  }
          #}}
        ]
      ).to_a
    end

    def is_lessor
      false
    end

    # shortcut d'association
    def main_contact_id=(id)
      contact_group&.set(main_contact_id: id, updated_at: DateTime.current)
    end

    # shortcut d'accès au group
    def supervisor_group
      Group.find_by(id: supervisor_group_id)
    end

    # shortcut d'accès au group
    def manager_group
      Group.find_by(id: manager_group_id)
    end

    # déclenche une erreur liée à la parenté
    def trigger_parenting_error
      raise Exception, "entity #{id} isn't root but has no parent"
    end

    # récupère la version de l'instance en passant par le modèle Hermes::Lessor
    def lessor_instance
      Hermes::Lessor.find_by(id: id)
    end

    # cherche l'élémént le plus haut dans l'arborescence
    def find_root(target_class_name = 'entity', return_iteration = false)
      return self if is_root == true
      return parent if parent.is_root == true

      found_grand_parent = parent&.parent
      parent.trigger_parenting_error unless found_grand_parent.present?

      step = 0
      intermediate_root = found_grand_parent
      while intermediate_root.is_root != true
        if (step_target = intermediate_root.parent).present?
          intermediate_root = step_target
        else
          intermediate_root.trigger_parenting_error
        end
        step += 1
      end
      return unless intermediate_root

      found_root = case target_class_name
                   when 'entity'
                     intermediate_root
                   when 'lessor'
                     intermediate_root.lessor_instance
                   end
      if return_iteration == true
        { object: found_root, iteration_quantity: step }
      else
        found_root
      end
    end

    # =========================
    # SIDE EFFECTS DEPENDENCIES
    # =========================

    # gestionnaire du stockage de l'id de l'élément le plus haut dans l'arborescence
    def handle_root_id
      self.root_id = is_root ? id : parent.root_id
    end

    # gestionnaire du booléen identifiant si l'objet est à la racine
    def handle_is_root
      self.is_root = !parent_id.present?
    end

    # gestionnaire de mise à jour du total des points dépensés et la balance
    def update_spent_points
      self.total_spent = SPENT_ON_FIELDS.inject(0) do |sum, field|
        sum + (send(field) || 0)
      end
      self.available_points = (annual_stock || 0) - (total_spent || 0)
    end

    # gestionnaire de génération des groupes essentiels aux ACL
    def create_and_set_own_groups
      group = Group.find_or_create_by(name: "entity_#{id}")
      self.group_id = group.id
      supervisor_group = Group.find_or_create_by(name: "entity_supervisor_#{id}")
      self.supervisor_group_id = supervisor_group.id
      manager_group = Group.find_or_create_by(name: "entity_manager_#{id}")
      self.manager_group_id = manager_group.id
      readers.concat(["g_#{group.id}"])
      writers.concat(["g_#{manager_group.id}", "g_#{supervisor_group.id}"])
      [supervisor_group, manager_group, group].each do |target_group|
        target_group.set(
          writers: (target_group.writers + [user_id]).uniq.compact,
          updated_at: DateTime.current
        )
      end
    end

    # =====================
    # SIDE EFFECTS HANDLERS
    # =====================

    def before_create_actions
      handle_is_root
      handle_root_id
      create_and_set_own_groups
    end

    def before_actions
      handle_is_root
      handle_root_id
    end

    def before_save_actions
      update_siren
      clean_integer
      update_spent_points if NOT_NULL_INTEGER_FIELDS.any? { |field| send("#{field}_changed?") }
    end

    def clean_integer
      NOT_NULL_INTEGER_FIELDS.each do |field|
        self[field] = 0 if send(field).nil?
      end 
    end 
    
    def update_siren
      self.siren = (siren.blank?) ? siret&.slice(0, 9) :  siren&.slice(0, 9)
    end

    def after_actions
      check_status_changed
      
      if root_id_changed? && root_id_was && root_id && is_root
        Hermes::Entity.where(root_id: root_id_was).set(root_id: root_id, updated_at: DateTime.now)
      end

      AutoRequestValidationChangedJob.perform_later(id.to_s) if auto_request_validation && auto_request_validation_changed?

      if siret_changed? || siren_changed?
        HousingRequestEntityUpdateJob.perform_later(id.to_s, siret_was, siren_was)
      end 
    end

    def after_destroy_actions
      # ! ATTENTION !
      # peut potentiellement causer des problématiques de lock
      direct_children.each do
        child.handle_root_id
        child.save
      end
    end


    def update_siren
      self.siren = (siren.blank?) ? siret&.slice(0, 9) :  siren&.slice(0, 9)
    end

    def switch_manual_mode
      if self.manual_mode == false 
        self.manual_mode = true 
      else 
        self.manual_mode = false 
      end
      self.save 
    end 

    def create_entity_request_account(points, housing_request)
      available_points = self.available_points
      if points <= available_points
        if entity_request = Hermes::EntityRequestAccount.create({
                                                                'amount' => points,
                                                                'origin_entity_id' => self.id,
                                                                'entity_id' => self.id})
          housing_request.entity_request_account << entity_request
  
          return {status: "success", code: 200, message: {'message' => 'points successfully added', 'points_added' => points}}
        else 
          return {status: "internal servor error", code: 405, message: 'couldn\'t create a entity request please contact support'}
        end                                                                                
      else
        parent_entity = self.parent

        if !parent_entity.nil?
          parent_points = parent_entity.available_points
          final_points = available_points - points
          if final_points <= parent_points
            if entity_request = Hermes::EntityRequestAccount.create({
                                                                'amount' => available_points,
                                                                'origin_entity_id' => self.id,
                                                                'entity_id' => self.id})
              housing_request.entity_request_account << entity_request
              if parent_entity_request = Hermes::EntityRequestAccount.create({
                                                                              'amount' => final_points,
                                                                              'origin_entity_id' => self.id,
                                                                              'entity_id' => parent_entity.id})
                housing_request.entity_request_account << parent_entity_reques
                return {status: "success", code: 200, message: {'message' => 'points successfully added', 'points_added' => points}}
              else 
                return {status: "internal servor error", code: 405, message: 'couldn\'t create a entity request please contact support'}
              end 
            else
              return {status: "internal servor error", code: 405, message: 'couldn\'t create a entity request please contact support'}
            end
          else 
            return {status: "method unauthorized", code: 405, message: 'Too much points required not enough points on entity and entity parent'}
          end 
        else 
          return {status: "method unauthorized", code: 405, message: 'Too much points required not enough points on entity and no entity parent'}
        end 
      end
    end 

    def user_has_user_right?(user, type = 'entity')
      id_field = type == 'entity' ? :entity_id : :lessor_id
      Hermes::UserRight.where(id_field => self.id, user_id: user.id).exists?
    end
  
    # =============
    # CLASS METHODS
    # =============
    class << self
      # utilitaire d'aggrégation d'arborescence
      def rex_aggregation(
        root_element_id,
        target_class_name = 'entity',
        max_depth = nil,
        for_user = nil,
        use_acl = true
      )
        return unless root_element_id.present?

        target_klass = target_class(target_class_name)

        # https://docs.mongodb.com/manual/reference/operator/aggregation/graphLookup/
        # mongo version >= 3.4
        graphlookup_hash = {
          from: target_klass.collection.name,
          startWith: '$_id',
          connectFromField: '_id',
          connectToField: 'parent_id',
          as: 'children',
          depthField: 'query_depth'
        }
        graphlookup_hash[:maxDepth] = max_depth if max_depth
        if use_acl == true
          acl_selector = target_klass.readable_by(for_user).selector
          # restrict_query = acl_selector.clone.merge({ root_id: root_element_id })
          # unecessary
          graphlookup_hash[:restrictSearchWithMatch] = acl_selector
        end

        target_klass.collection.aggregate([
          { :$match => { '_id': BSON::ObjectId(root_element_id) } },
          { :$graphLookup => graphlookup_hash }
        ])
      end

      # utilitaire d'aggregation d'arborescence avec plus d'options en paramètre
      def tree_aggregation(
        target_id,
        from_parent_depth = 0,
        depth_from_target = -1,
        target_class_name = 'entity',
        for_user = nil,
        use_acl = false
      )
        # @target_id:         id de l'objet depuis lequel produire la recherche
        # @from_parent_depth: niveau du parent depuis lequel générer l'arborescence ; -1 == root
        # @depth_from_target: niveau de profondeur, depuis le parent, jusqu'à laquelle générer l'arborescence ; -1 == profondeur maximale
        # @target_class_name: restreint le retour au type de model choisit
        raise ArgumentError, "from_parent_depth parameter must be '>= 0' || '== -1'" unless from_parent_depth == -1 || from_parent_depth >= 0

        target = Hermes::Entity.find_by(id: target_id)
        from_parent = if from_parent_depth == 0
                        target
                      elsif from_parent_depth == -1 || from_parent_depth > 0
                        method_result = target.parent(from_parent_depth, target_class_name, true)
                        if method_result.blank?
                          from_parent_depth = method_result[:iteration_quantity]
                          method_result[:object]
                        end
                      end
        raise ArgumentError, "no parent found at depth #{from_parent_depth}" if from_parent.blank?

        seeked_depth = if depth_from_target.negative?
                         nil
                       else
                         from_parent_depth + depth_from_target
                       end

        rex_aggregation(from_parent.id, target_class_name, seeked_depth, for_user, use_acl)
      end
    end
  end
end
