require 'rails_helper'
require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

RSpec.describe BatchTasks::NewReservedOfferNotification, type: :service do

  subject { described_class.new }

  describe 'create_new_reserved_offers_notifications' do 
    let!(:user) {create(:user)}
    let!(:entity_1) {create(:entity, siret: '00000000000001')}
    let!(:offer1) {create(:published_offer, :with_eligible_requests, requests_count: 1)}
    let!(:offer2) {create(:published_offer, :with_eligible_requests, requests_count: 2, external_ref: 'offer2')}
    let!(:hr1) {offer1.eligibles_requests.first}
    let!(:entity_reservation1) { create(:entity_reservation, entity: entity_1, housing_offer: offer1, from: DateTime.now() - 20.days, to: DateTime.now() + 30.days )}
    let!(:entity_reservation2) { create(:entity_reservation, entity: entity_1, housing_offer: offer2, from: DateTime.now() - 20.days, to: DateTime.now() + 30.days )}

    before do 
      create_mail_template_default()
      hr1.main_applicant.update(main_employer_siret: entity_1.siret)
      hr1.set(entity_validated: true, request_status: 'COMPLETE', user_id: user.id)
    end

    it 'works' do 
      expect(hr1.reload.entity_id.to_s).to eq(entity_1.id.to_s)
      expect(hr1.reload.eligibles_offers.count).to eq(2)
      
      expect do
        subject.generate_and_enqueue_notification(
          date_start: DateTime.now - 2.days, 
          date_end: DateTime.now + 1.minutes,
        )
      end.to change { enqueued_jobs.select{|job| job[:job] == SlowNotifierJob }.length }.by 1

      async_notif_job = execute_last_job(SlowNotifierJob)
      expect(async_notif_job[:queue]).to eq('slow_notifications')
      expect(async_notif_job[:args][0]["datas"]["offer_ids"].sort).to eq([offer1.id.to_s, offer2.id.to_s].sort)

      email_job = execute_last_job(ActionMailer::DeliveryJob) 
      expect(email_job[:queue]).to eq('slow_notifications')
    end 

    it 'with offer_filter' do
      expect do
        subject.generate_and_enqueue_notification(
          date_start: DateTime.now - 2.days, 
          date_end: DateTime.now + 1.minutes,
          offer_filter: {external_ref: offer2.external_ref.to_s}
        )
      end.to change { enqueued_jobs.select{|job| job[:job] == SlowNotifierJob }.length }.by 1

      last_job = execute_last_job(SlowNotifierJob)
      expect(last_job[:args][0]["datas"]["offer_ids"].sort).to eq([offer2.id.to_s])
    end 
  end
end